"javascript specific code for stage 3."

import glob
import json
import os
import sys

from depends_on.common import log


def load_package_json(package_json_path):
    """Load package.json file."""
    with open(package_json_path, "r", encoding="UTF-8") as package_json:
        return json.load(package_json)


def local_dependencies(dirs):
    """Return a dictionary of local dependencies with the name as key
    and the directory as value. It also looks for workspaces in package.json."""
    deps = {}
    for local_dir in dirs:
        if "subdir" in dirs[local_dir]:
            local_path = os.path.join(
                dirs[local_dir]["path"], dirs[local_dir]["subdir"]
            )
        else:
            local_path = dirs[local_dir]["path"]
        package_json_path = os.path.join(local_path, "package.json")
        if os.path.exists(package_json_path):
            package = load_package_json(package_json_path)
            deps[package["name"]] = dirs[local_dir]
            if "workspaces" in package and "packages" in package["workspaces"]:
                log("detected workspaces")
                for workspace_glob in package["workspaces"]["packages"]:
                    log(f"processing {workspace_glob}")
                    for workspace_dir in glob.glob(
                        os.path.join(local_path, workspace_glob)
                    ):
                        log(f"found subdir={workspace_dir}")
                        workspace_path = os.path.join(local_path, workspace_dir)
                        log(f"checking workspace {workspace_path} for package.json")
                        workspace_package_json_path = os.path.join(
                            workspace_path, "package.json"
                        )
                        if os.path.exists(workspace_package_json_path):
                            workspace_package = load_package_json(
                                workspace_package_json_path
                            )
                            log(f"found package {workspace_package['name']}")
                            deps[workspace_package["name"]] = {
                                "path": workspace_path,
                                "subdir": workspace_dir[len(local_path) + 1 :],
                                "fork_url": dirs[local_dir]["fork_url"],
                                "branch": dirs[local_dir]["branch"],
                            }
    return deps


def process_dependencies(dependencies, dirs, container_mode, package_json_path):
    """Process dependencies in package.json and replace local dependencies"""
    local_deps = local_dependencies(dirs)
    log(f"Found {len(local_deps)} local dependencies: {local_deps=}")
    count = 0
    for dependency in dependencies:
        if dependency in local_deps:
            if container_mode:
                info = local_deps[dependency]
                repo = "git+" + info["fork_url"] + "#" + info["branch"]
                log(
                    f"Replacing {dependency} with remote version from {repo} in package.json"
                )
                dependencies[dependency] = repo
            else:
                log(
                    f"Replacing {dependency} with local version from {local_deps[dependency]['path']} in package.json"
                )
                dependencies[dependency] = "file:" + local_deps[dependency]["path"]
            count += 1
    return count


def process_javascript(main_dir, dirs, container_mode):
    """Use changes from PR in package.json if present"""
    package_json_path = os.path.join(main_dir, "package.json")
    if not os.path.exists(package_json_path):
        return False
    log("Detected package.json file, checking for local dependencies")
    package = load_package_json(package_json_path)
    if "dependencies" not in package:
        return False
    dependencies = package["dependencies"]
    count = process_dependencies(dependencies, dirs, container_mode, package_json_path)
    if "devDependencies" in package:
        dependencies = package["devDependencies"]
        count += process_dependencies(
            dependencies, dirs, container_mode, package_json_path
        )
    with open(package_json_path, "w", encoding="UTF-8") as package_json:
        json.dump(package, package_json, indent=2)
    return count > 0


# javascript.py ends here



"Python specific code for stage 3."

import os
import re
import sys

from depends_on.common import log


def lookup_name(fname):
    "Lookup the name of a module"
    if os.path.exists(fname):
        log(f"Looking up name in {fname}")
        with open(fname, "r", encoding="UTF-8") as in_stream:
            for line in in_stream.readlines():
                match = re.match(r"^\s*name\s*=\s*['\"](.*?)['\"]\s*,", line)
                if match:
                    return match.group(1)
    return None


def get_modules(dirs):
    "Get the dictionary of python modules from the local dependencies."
    python_mods = {}
    for _dir in dirs:
        name = lookup_name(os.path.join(dirs[_dir]["path"], "setup.py")) or lookup_name(
            os.path.join(dirs[_dir]["path"], "pyproject.toml")
        )
        if name:
            python_mods[name] = dirs[_dir]
    return python_mods


def process_python_requirements(main_dir, dirs, container_mode):
    "Replace modules in requirements.txt for the local dependencies."
    requirements_txt = os.path.join(main_dir, "requirements.txt")
    requirements_txt_new = requirements_txt + ".new"
    if not os.path.exists(requirements_txt):
        return False
    log("requirements.txt detected")
    module_dirs = get_modules(dirs)
    log(f"{module_dirs=}")
    # replace the modules in requirements.txt
    nb_replace = 0
    with open(requirements_txt, "r", encoding="UTF-8") as in_stream:
        with open(requirements_txt_new, "w", encoding="UTF-8") as out_stream:
            for line in in_stream.readlines():
                match = re.match(r"^\s*(\w+)", line)
                if match and match.group(1) in module_dirs:
                    mod = match.group(1)
                    if container_mode:
                        # doc at https://pip.pypa.io/en/stable/cli/pip_install/#git
                        pkg = f"{mod} @ git+{module_dirs[mod]['fork_url']}@{module_dirs[mod]['branch']}"
                        if "subdir" in module_dirs[mod]:
                            pkg += (
                                f"#egg=subdir&subdirectory={module_dirs[mod]['subdir']}"
                            )
                        log(f"Replacing {mod} in requirements.txt with {pkg}")
                        out_stream.write(f"{pkg}\n")
                    else:
                        log(
                            f"Replacing {mod} in requirements.txt with {module_dirs[mod]['path']}"
                        )
                        out_stream.write(f"-e {module_dirs[mod]['path']}\n")
                    nb_replace += 1
                else:
                    out_stream.write(line)
    os.rename(requirements_txt_new, requirements_txt)
    return nb_replace > 0


def process_python_pyproject(main_dir, dirs, container_mode):
    "Replace modules in pyproject.toml for the local dependencies."
    pyproject_toml = os.path.join(main_dir, "pyproject.toml")
    pyproject_toml_new = pyproject_toml + ".new"
    if not os.path.exists(pyproject_toml):
        return False
    log("pyproject.toml detected")
    # get the list of python packages from local dependencies
    module_dirs = get_modules(dirs)
    log(f"{module_dirs=}")
    # replace the modules in pyproject.toml
    nb_replace = 0
    with open(pyproject_toml, "r", encoding="UTF-8") as in_stream:
        with open(pyproject_toml_new, "w", encoding="UTF-8") as out_stream:
            for line in in_stream.readlines():
                match = re.match(r"^\s*(\w+)\s*=", line)
                if match and match.group(1) in module_dirs:
                    mod = match.group(1)
                    if container_mode:
                        # doc at https://python-poetry.org/docs/dependency-specification/#git-dependencies
                        pkg = f"{mod} = {{ git = \"{module_dirs[mod]['fork_url']}\", branch = \"{module_dirs[mod]['branch']}\""
                        if "subdir" in module_dirs[mod]:
                            pkg += (
                                f", subdirectory = \"{module_dirs[mod]['subdir']}\" }}"
                            )
                        else:
                            pkg += " }"
                        log(f"Replacing {mod} in pyproject.toml with {pkg}")
                        out_stream.write(f"{pkg}\n")
                    else:
                        log(
                            f"Replacing {mod} in pyproject.toml with {module_dirs[mod]['path']}"
                        )
                        out_stream.write(
                            f'{mod} = {{ path = "{module_dirs[mod]["path"]}" }}\n'
                        )
                    nb_replace += 1
                else:
                    out_stream.write(line)
    os.rename(pyproject_toml_new, pyproject_toml)
    return nb_replace > 0


def process_python(main_dir, dirs, container_mode):
    "Process python dependencies."
    # process pyprohect.toml first because they can both be present
    # and it supposed to be the main one
    return process_python_pyproject(
        main_dir, dirs, container_mode
    ) or process_python_requirements(main_dir, dirs, container_mode)


# python.py ends here



"Functions used by multiple stages."

import json
import os
import re
import sys
import urllib.parse
from urllib.request import Request, urlopen

_SENSITIVE_STRINGS = []


def add_sensitive_string(string):
    "Add a string to the list of sensitive strings."
    if string and string != "" and string not in _SENSITIVE_STRINGS:
        _SENSITIVE_STRINGS.append(string)


def init_sensitive_strings():
    "Initialize the list of sensitive strings."
    for env_var in ("GITHUB_TOKEN", "GITLAB_TOKEN"):
        add_sensitive_string(os.environ.get(env_var))


def log(message):
    "Log a message to stderr after masking sensitive strings."
    for sensitive_string in _SENSITIVE_STRINGS:
        message = message.replace(sensitive_string, "***")
    print(message, file=sys.stderr)


def get_json_url(url, **headers):
    "Get the content of an URL."
    req = Request(url)
    for header, value in headers.items():
        req.add_header(header, value)
    with urlopen(req) as response:
        response_content = response.read()
    response_content.decode("utf-8")
    return json.loads(response_content)


def save_depends_on(data, dirname):
    "Save the data to a JSON file."
    with open(
        os.path.join(dirname, "depends-on.json"),
        "w",
        encoding="UTF-8",
    ) as json_stream:
        json.dump(data, json_stream, indent=2)


def get_pull_request_info(org, repo, pr_number):
    "Get the information about a GitHub Pull request."
    token = os.environ.get("GITHUB_TOKEN")
    # get the information about the Pull request using the GitHub API
    # set the Authorization header to use the token
    headers = {"Accept": "application/vnd.github.v3+json"}
    if token:
        log("Using GitHub token")
        headers["Authorization"] = f"token {token}"
    pr_info = get_json_url(
        f"https://api.github.com/repos/{org}/{repo}/pulls/{pr_number}", **headers
    )
    return pr_info


def get_gerrit_change_info(gerrit_url, gerrit_change_id):
    "Get the information about the Gerrit change."
    req = Request(
        f"{gerrit_url}/changes/{gerrit_change_id}?o=CURRENT_REVISION&o=CURRENT_COMMIT",
    )
    req.add_header("Accept", "application/json")
    with urlopen(req) as response:
        response_content = response.read()
    # remove the magic prefix
    response_content = re.sub(r"^\)\]\}\'\n", "", response_content.decode("utf-8"))
    change_info = json.loads(response_content)
    return change_info


def extract_pull_request(depends_on_url, check_mode):
    "Extract the dependency by git cloning the repository in the right branch for the Pull request."
    # parse the URL to extract the repo, org and pr number
    # the format is https://github.com/<org>/<repo>/pull/<pr_number>?subdir=<subdir>&<key>=<value>
    url_parts = depends_on_url.split("/")
    if len(url_parts) != 7:
        raise ValueError(f"Invalid URL {depends_on_url}")
    org = url_parts[3]
    repo = url_parts[4]
    pr_data = url_parts[6].split("?", 1)
    pr_number = pr_data[0]
    if len(pr_data) == 2:
        pr_data = pr_data[1].split("&")
        pr_data = {x.split("=")[0]: x.split("=")[1] for x in pr_data}
    else:
        pr_data = {}
    pr_info = get_pull_request_info(org, repo, pr_number)
    top_dir = os.path.realpath(repo)

    if not check_mode:
        # extract the dependency on disk
        extract_github_change(
            pr_info["head"]["repo"]["clone_url"],
            pr_info["head"]["ref"],
            pr_info["base"]["repo"]["clone_url"],
            pr_info["base"]["ref"],
            repo,
        )

        # save the information about the Pull request in depends-on.json
        data = {
            "description": pr_info["body"],
            "fork_url": pr_info["head"]["repo"]["clone_url"],
            "branch": pr_info["head"]["ref"],
            "main_url": pr_info["base"]["repo"]["clone_url"],
            "main_branch": pr_info["base"]["ref"],
            "topdir": top_dir,
            "path": top_dir,
            "merged": pr_info["merged"],
        }
        if "subdir" in pr_data:
            data["subdir"] = pr_data["subdir"]
            data["path"] = os.path.join(top_dir, pr_data["subdir"])
        save_depends_on(data, repo)
        log(f"PR data: {data}")
    return pr_info["merged"], top_dir


def extract_gerrit_review(depends_on_url, check_mode):
    "Extract the dependency by git cloning the repository in the right branch for the Gerrit change."
    # Parse the URL to extract the url and change number.
    # The format is
    # https://gerrit.wikimedia.org/r/c/mediawiki/extensions/ContentTranslation/+/123456
    # to extract https://gerrit.wikimedia.org/r and 123456
    url_parts = depends_on_url.split("/c/")
    if len(url_parts) != 2:
        raise ValueError(f"Invalid URL {depends_on_url}")
    gerrit_url = url_parts[0]
    change_parts = url_parts[1].split("/")
    change_id = change_parts[-1]
    # Get the information about the Gerrit change
    change_info = get_gerrit_change_info(gerrit_url, change_id)
    project = os.path.basename(change_info["project"])
    top_dir = os.path.realpath(project)
    if not check_mode:
        # extract the dependency on disk
        extract_gerrit_change(
            change_info["revisions"][change_info["current_revision"]]["fetch"][
                "anonymous http"
            ]["url"],
            change_info["revisions"][change_info["current_revision"]]["fetch"][
                "anonymous http"
            ]["ref"],
            change_info["branch"],
            project,
        )
        # save the information about the Gerrit change in depends-on.json
        data = {
            "description": change_info["revisions"][change_info["current_revision"]][
                "commit"
            ]["message"],
            "fork_url": change_info["revisions"][change_info["current_revision"]][
                "fetch"
            ]["anonymous http"]["url"],
            "branch": change_info["revisions"][change_info["current_revision"]][
                "fetch"
            ]["anonymous http"]["ref"],
            "main_url": change_info["revisions"][change_info["current_revision"]][
                "fetch"
            ]["anonymous http"]["url"],
            "main_branch": change_info["branch"],
            "topdir": top_dir,
            "path": top_dir,
            "merged": change_info["status"] == "MERGED",
        }
        save_depends_on(data, project)
        log(f"Change data: {data}")
    return change_info["status"] == "MERGED", top_dir


def get_gitlab_project_info(gitlab_url, project, headers):
    "Get the project id from the project path"
    # The format of the project path is /<org>/<project>
    # We need to replace / by %2F
    project = project.replace("/", "%2F")
    # Get the project id from the API
    url = f"{gitlab_url}/api/v4/projects/{project}"
    return get_json_url(url, **headers)


def get_gitlab_auth():
    "Get the authentication part of the URL"
    if "GITLAB_USER" in os.environ and "GITLAB_TOKEN" in os.environ:
        return f"{os.environ['GITLAB_USER']}:{os.environ['GITLAB_TOKEN']}@"
    if "GITLAB_TOKEN" in os.environ:
        return f"{os.environ['GITLAB_TOKEN']}@"
    return ""


def extract_gitlab_merge_request(depends_on_url, check_mode):
    "Extract a gitlab merge request"
    # Parse the URL to extract the project and merge request number.
    # The format is https://<server>/<project>/-/merge_requests/<mr_number>
    # to extract https://<server> <project> and, <mr_number>
    url_parts = depends_on_url.split("/-/merge_requests/")
    if len(url_parts) != 2:
        raise ValueError(f"Invalid URL {depends_on_url}")
    mr_number = url_parts[1]
    url_parts = urllib.parse.urlparse(url_parts[0])
    gitlab_url = f"{url_parts.scheme}://{url_parts.netloc}"
    # if authentication is needed for the gitlab server:
    # - for the API we need to add the PRIVATE-TOKEN header
    # - for git, the authentication is part of the URL https://<username>:<token>@<host>/...
    auth_info = get_gitlab_auth()
    auth_gitlab_url = f"{url_parts.scheme}://{auth_info}{url_parts.netloc}"
    project = url_parts.path
    if "GITLAB_TOKEN" in os.environ:
        headers = {"PRIVATE-TOKEN": os.environ["GITLAB_TOKEN"]}
    else:
        headers = {}
    project_info = get_gitlab_project_info(gitlab_url, project[1:], headers)
    project_id = project_info["id"]
    # get the information about the merge request
    url = f"{gitlab_url}/api/v4/projects/{project_id}/merge_requests/{mr_number}"
    mr_info = get_json_url(url, **headers)
    source_project_info = get_gitlab_project_info(
        gitlab_url, str(mr_info["source_project_id"]), headers
    )
    source_url = source_project_info["http_url_to_repo"].replace(
        "://", "://" + auth_info, 1
    )
    base_project = os.path.basename(project)
    top_dir = os.path.realpath(base_project)
    if not check_mode:
        extract_gitlab_change(
            f"{auth_gitlab_url}{project}.git",
            f"{source_url}",
            mr_info["source_branch"],
            mr_info["target_branch"],
            base_project,
        )
        # save the information about the merge request in depends-on.json
        data = {
            "description": mr_info["description"],
            "fork_url": source_url,
            "branch": mr_info["source_branch"],
            "main_url": f"{gitlab_url}{project}.git",
            "main_branch": mr_info["target_branch"],
            "topdir": top_dir,
            "path": top_dir,
            "merged": mr_info["state"] == "merged",
        }
        save_depends_on(data, base_project)
        log(f"Change data: {data}")
    return mr_info["state"] == "merged", top_dir


def extract_depends_on(depends_on_url, check_mode):
    "Extract the dependency by git cloning the repository in the right branch."
    if "/c/" in depends_on_url:
        return extract_gerrit_review(depends_on_url, check_mode)
    elif "gitlab" in urllib.parse.urlparse(depends_on_url).netloc:
        return extract_gitlab_merge_request(depends_on_url, check_mode)
    else:
        return extract_pull_request(depends_on_url, check_mode)


def extract_gitlab_change(base_url, change_url, branch, main_branch, repo):
    "Extract the dependency by git cloning the repository in the right branch for the Merge request."
    command(f"git clone --filter=tree:0 {base_url} {repo}")
    command(
        f"cd {repo} && git fetch {change_url} {branch} && git checkout -b {branch} FETCH_HEAD"
    )
    merge_main_branch(repo, main_branch)
    return repo


def extract_github_change(fork_url, branch, main_url, main_branch, repo):
    "Extract the dependency by git cloning the repository in the right branch for the Pull request."
    command(f"git clone --filter=tree:0 {main_url} {repo}")
    command(f"cd {repo} && git remote add pr {fork_url} && git fetch pr {branch}")
    # extract the master/main branch name
    command(f"cd {repo} && git checkout -b {branch} --track pr/{branch}")
    merge_main_branch(repo, main_branch)
    return repo


def extract_gerrit_change(change_url, branch, main_branch, repo):
    "Extract the dependency by git cloning the repository in the right branch for the Gerrit change."
    command(f"git clone --filter=tree:0 {change_url} {repo}")
    command(
        f"cd {repo} && git fetch {change_url} {branch} && git checkout -b {branch} FETCH_HEAD"
    )
    merge_main_branch(repo, main_branch)
    return repo


def merge_main_branch(repo, main_branch):
    "Merge the main branch into the current branch."
    # set a dummy user name and email for the merge process to work
    command(
        f"cd {repo} && git config user.name 'Depends-On' && git config user.email 'depends-on@localhost' && git config commit.gpgsign false"
    )
    # merge the main branch into the current branch
    command(f"cd {repo} && git merge origin/{main_branch} --no-edit")
    return repo


def check_error(status, message):
    "Check the status and exit if it is false."
    if not status:
        raise Exception(message)


def command(cmd):
    "Execute a command"
    log(f"+ {cmd}")
    ret = os.system(cmd)
    check_error(ret == 0, f"Command failed with exit code {ret}")


# common.py ends here



import json

from setuptools import setup

setup(
    name="depends-on",
    description="A Python library to manage dependencies between changes.",
    long_description_content_type="text/markdown",
    long_description=open("README.md").read(),
    author="The Depends-On Team",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Software Development :: Version Control :: Git",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Quality Assurance",
    ],
    version=json.load(open("package.json"))["version"],
    packages=["depends_on"],
    scripts=["depends_on_stage1", "depends_on_stage2", "depends_on_stage3"],
)
